use cpython::{Python, PyErr, PyModule, PyObject, PyResult, PyString, PyTuple};
use cpython::{FromPyObject, NoArgs, PythonObject, ToPyObject};
use cpython::exc::{ValueError};
use std::str;

pub fn initialize_module(py: Python, m: &PyModule) -> PyResult<()> {
    m.add(py, "_shorten_step", py_fn!(py, py_shorten_step(word: String) -> PyResult<String> {
        let mut result = word.into_bytes();
        shorten_step(&mut result);
        Ok(String::from_utf8(result).unwrap())
    }))?;
    m.add(py, "_shorten", py_fn!(py, py_shorten(word: String) -> PyResult<String> {
        let mut result = word.into_bytes();
        shorten(&mut result);
        Ok(String::from_utf8(result).unwrap())
    }))?;
    m.add(py, "_normalize", py_fn!(py, py_normalize(word: String) -> PyResult<Coordinate> {
        let mut result = word.into_bytes();
        normalize(&mut result);
        Ok(Coordinate(result))
    }))?;
    m.add(py, "neighbors", py_fn!(py, py_neighbors(tile: Coordinate) -> PyResult<PyTuple> {
        let neighbors = neighbors(&tile);
        let py_neighbors: Vec<PyObject> = neighbors.iter().map(
            |obj| obj.to_py_object(py).into_object()
        ).collect();
        Ok(PyTuple::new(py, &py_neighbors))
    }))?;
    m.add(py, "neighbor_index", py_fn!(py,
        py_neighbor_index(tile: Coordinate, neighbor: Coordinate) -> PyResult<usize> {
            if let Some(index) = neighbor_index(&tile, &neighbor) {
                Ok(index)
            } else {
                Err(PyErr::new::<ValueError, NoArgs>(py, NoArgs))
            }
        }))?;
    Ok(())
}

// We want to express coordinates as specific words generated by the hyper steps.
// The hyper steps are 'x' and 'y', and you perform multiple steps by concatenating.
// For group theoretical reasons, steps are perfomed "backwards" to what you expect:
// "xy" first applies 'y', then 'x'.
type Word = Vec<u8>;

// A coordinate for a square is the least word that represents it.
// The ordering we use is shortlex: first by length, then lexicographic.
//
// To be exact, a coordinate of a square is the minimum of all words
// that transform the origin to a rotation of the given square.
custom_derive! {
    #[derive(Clone, Debug, NewtypeFrom, PartialEq)]
    pub struct Coordinate(Word);
}

impl Coordinate {
    pub fn to_word(&self) -> &Word {
        let &Coordinate(ref coord) = self;
        coord
    }
}

impl<'source> FromPyObject<'source> for Coordinate {
    fn extract(py: Python, obj: &'source PyObject) -> PyResult<Self> {
        let string = String::extract(py, obj)?;
        Ok(Coordinate(string.into_bytes()))
    }
}
impl ToPyObject for Coordinate {
    type ObjectType = PyString;
    fn to_py_object(&self, py: Python) -> Self::ObjectType {
        let coord = self.to_word();
        str::from_utf8(coord).map(String::from).unwrap().to_py_object(py)
    }
}

// Is the first word less than or equal to the second word by shortlex?
fn leq(w1: &Word, w2: &Word) -> bool {
    if w1.len() == w2.len() {
        w1 <= w2
    } else {
        w1.len() <= w2.len()
    }
}

// Replace a given part of the word with another, in-place.
// Returns whether any replacement was performed.
fn replace(word: &mut Word, from: &[u8], to: &[u8]) -> bool {
    assert!(from.len() >= to.len());
    let mut found = false;
    let mut write_index = 0;
    while write_index + from.len() <= word.len() {
        if from == &word[write_index .. write_index + from.len()] {
            found = true;

            // The index to the part of the word we didn't replace.
            let mut read_index = write_index + from.len();

            // First copy the replacement.
            let mut to_index = 0;
            while to_index < to.len() {
                word[write_index] = to[to_index];
                write_index += 1;
                to_index += 1;
            }

            // Then move over the remainder of the word.
            if read_index > write_index {
                while read_index < word.len() {
                    word[write_index] = word[read_index];
                    write_index += 1;
                    read_index += 1;
                    assert!(read_index > write_index);
                }
                word.truncate(write_index);
            } else {
                assert!(read_index == write_index);
            }
        }
        write_index += 1;
    }
    found
}

// The Knuth-Bendix algorithm gives a rewrite system
// consisting of these steps applied repeatedly.
// This function is exposed so it can be tested.
// Returns whether the word was shortened.
fn shorten_step(word: &mut Word) -> bool {
    replace(word, "yxy".as_bytes(), "xxx".as_bytes()) ||
    replace(word, "xxxx".as_bytes(), "".as_bytes()) ||
    replace(word, "yyyy".as_bytes(), "xyx".as_bytes()) ||
    replace(word, "xxxyyy".as_bytes(), "yxxyx".as_bytes()) ||
    replace(word, "xyxxyx".as_bytes(), "yyy".as_bytes()) ||
    replace(word, "yyyxxx".as_bytes(), "xyxxy".as_bytes()) ||
    replace(word, "xxxyyxxx".as_bytes(), "yxxyxxy".as_bytes()) ||
    replace(word, "xyxxyyxxyx".as_bytes(), "yyyxxyyy".as_bytes()) ||
    replace(word, "xyxxyyxxyyxxyx".as_bytes(), "yyyxxyyyxxyyy".as_bytes()) ||
    replace(word, "yyyxxyyyxxyyyxxyyy".as_bytes(), "xyxxyyxxyyxxyyxxyx".as_bytes()) ||
    replace(word, "yxxyxxyyxxyyxxyyxxyx".as_bytes(), "xxxyyxxyyyxxyyyxxyyy".as_bytes()) ||
    replace(word, "yyyxxyyyxxyyyxxyyxxx".as_bytes(), "xyxxyyxxyyxxyyxxyxxy".as_bytes())
}

// Internal function for making sure a word is in the unique-per-word representation.
// You shouldn't use this for making coordinates as it does not consider rotations
// of the same word to be equal.
fn shorten(word : &mut Word) {
    while shorten_step(word) { }
}

// Internal function for making sure a calculated word is the actual coordinate.
fn normalize(word: &mut Word) {
    shorten(word);
    let mut rotated = word.clone();
    // Calculate all rotations of the word and pick the smallest one.
    // Note that we already have considered the trivial rotation, so we start with 1.
    for _ in 1..4 {
        // Insert an 'x' rotation at the start
        // (which is at the end of the word, due to group theory and such)
        // giving us a transformed origin tile to perform all operations on.
        rotated.push(120);
        shorten(&mut rotated);
        if leq(&rotated, word) {
            word.clear();
            word.extend_from_slice(&rotated);
        }
    }
}

// The words for the squares of distance 1 to the origin.
//
// Note that these are words, not coordinates, as we preserve the orientation.

// COMPAT: should be the same as in coordinate/kbword.py
const NORTH: [u8; 4] = [121, 120, 120, 120];
const EAST: [u8; 4] = [120, 121, 120, 120];
const SOUTH: [u8; 4] = [120, 120, 121, 120];
const WEST: [u8; 4] = [120, 120, 120, 121];

// Transform the tile with the given word, giving a new tile.
fn concat(tile: &Coordinate, word: &[u8]) -> Coordinate {
    let &Coordinate(ref coord) = tile;
    let mut new_tile = coord.clone();
    new_tile.extend(word);
    normalize(&mut new_tile);
    Coordinate(new_tile)
}

// Give the four neighbors of a tile with given coordinates.
//
// The neighbors are listed in clockwise order.
// (i.e. from neighbors[0] to neighbors[1] you turn 90 degrees to the right.)
pub fn neighbors(tile: &Coordinate) -> [Coordinate; 4] {
    [
        concat(tile, &NORTH),
        concat(tile, &EAST),
        concat(tile, &SOUTH),
        concat(tile, &WEST)
    ]
}

// Give the index of the neighbor in the list of neighbors,
// or None if it's not a neighbor.
//
// The neighbors are listed in clockwise order.
// (i.e. from neighbors[0] to neighbors[1] you turn 90 degrees to the right.)
pub fn neighbor_index(tile: &Coordinate, neighbor: &Coordinate) -> Option<usize> {
    let neighborhood = neighbors(tile);
    for i in 0..neighborhood.len() {
        if neighborhood[i] == *neighbor {
            return Some(i)
        }
    }
    None
}

#[cfg(test)]
extern crate quickcheck;
#[cfg(test)]
mod tests {
    use coordinate;

    #[test]
    fn shorten_empty() {
        let mut word = Vec::new();
        coordinate::shorten(&mut word);
        assert!(word.len() == 0);
    }

    #[test]
    fn shorten_xxxx() {
        let mut word = String::from("xxxx").into_bytes();
        coordinate::shorten(&mut word);
        assert!(word.len() == 0);
    }

    #[test]
    fn shorten_xyxy() {
        let mut word = String::from("xyxy").into_bytes();
        coordinate::shorten(&mut word);
        assert!(word.len() == 0);
    }

    #[test]
    fn shorten_yxyx() {
        let mut word = String::from("yxyx").into_bytes();
        coordinate::shorten(&mut word);
        assert!(word.len() == 0);
    }

    #[quickcheck]
    fn shorten_shortens(mut word: coordinate::Word) {
        let len = word.len();
        coordinate::shorten(&mut word);
        assert!(len >= word.len());
    }
}
