import random

from dbecs.entity import Entity
from hyperbolic.block import Blocky
from hyperbolic.chunk import _chunk_size
from hyperbolic.coordinate import Coordinate
from hyperbolic.position import ChunkPosition, Positioned
from hyperbolic.size import Sized

class Region:
	"""Represents a group of chunks with the same appearance.
	
	Is used to generate the chunk's contents.
	"""

	def __init__(self):
		self.chunk_count = 0

	def generate_chunk(self, coordinate: Coordinate):
		"""Make a dictionary of the tiles in a chunk.
		
		Fills the chunk with some generic junk.
		"""
		self.chunk_count += 1
		return {(x, y, z): None
			for x in range(0, _chunk_size)
			for y in range(0, _chunk_size)
			for z in range(0, _chunk_size)
		}

# TODO: better name
def generate_background_foreground_decoration(coordinate, chunk, bg, fg, deco):
	# Add the ground, filling the full layer below the chunk
	ground = Entity()
	ground.add_component(Blocky, bg)
	ground.add_component(Positioned, coordinate, 0, 0, -1, 0)
	ground.add_component(Sized, _chunk_size, _chunk_size, 1)
	# Add foreground everywhere
	for x in range(0, _chunk_size):
		for y in range(0, _chunk_size):
			block = Entity()
			block.add_component(Blocky, fg)
			block.add_component(Positioned, coordinate, x, y, 1, 0)
			block.add_component(Sized, 1, 1, 1)
			# TODO: we must be able to do this better
			chunk[(x, y, 0)] = ground
			chunk[(x, y, 1)] = block

	flower_count = random.normalvariate(_chunk_size ** 2 / 10, 10)
	for _ in range(0, int(flower_count)):
		x = random.randrange(0, _chunk_size)
		y = random.randrange(0, _chunk_size)
		block = Entity()
		block.add_component(Blocky, deco)
		block.add_component(Positioned, coordinate, x, y, 1, 0)
		block.add_component(Sized, 1, 1, 1)
		chunk[(x, y, 1)] = block

	return chunk

class Desert(Region):
	"""A region with rocks and cactuses."""
	def generate_chunk(self, coordinate: Coordinate):
		chunk = super().generate_chunk(coordinate)
		return generate_background_foreground_decoration(coordinate, chunk, "sand", "rocks", "cactus")

class Field(Region):
	"""A region with grass and flowers."""
	def generate_chunk(self, coordinate: Coordinate):
		chunk = super().generate_chunk(coordinate)
		return generate_background_foreground_decoration(coordinate, chunk, "dirt", "grass", "flowers")

class Houses(Region):
	"""A region with a lot of houses."""
	
	def generate_chunk(self, coordinate: Coordinate):
		chunk = super().generate_chunk(coordinate)

		# Place down a building.
		building = """
#############
#     #     #
#     #     #
#     #     #
###  ##     #
#           #
#            
#            
#           #
#############
""".split('\n')
		height = len(building)
		width = max(map(len, building))
		offset_x = random.randrange(_chunk_size - width)
		offset_y = random.randrange(_chunk_size - height)
		for y, row in enumerate(building):
			for x, char in enumerate(row):
				if char == "#":
					block = Entity()
					block.add_component(Blocky, "bricks")
					block.add_component(Positioned, coordinate, x + offset_x, y + offset_y, 1, 0)
					block.add_component(Sized, 1, 1, 1)
					chunk[(x + offset_x, y + offset_y, 1)] = block

		return chunk

"""Contains all possible regions used to generate the world."""
region_options = [
	Desert,
	Field,
	Houses,
]
