import functools
import random

from dbecs.entity import Entity
from hyperbolic.block import Blocky
from hyperbolic.coordinate import origin
from hyperbolic.graphics import DrawQueue, position, rgb_color, white
from hyperbolic.input import keyboard_table
from hyperbolic.position import ChunkPosition, Positioned
from hyperbolic.size import Sized
from hyperbolic.world import World

def initialize(args):
	"""Start the simulation with given arguments.

	Returns an object that represents the current simulation state.
	"""
	world = World()

	return world

def update(world, dt):
	"""Update the simulation, with a given time step.

	Returns an object that represents the simulation state after updating.
	"""
	return world

@functools.lru_cache()
def get_pos(player, dx, dy):
	player_pos = Positioned(player)
	tile_pos = ChunkPosition(
			player_pos.coordinate,
			player_pos.local_x, player_pos.local_y, player_pos.local_z,
			player_pos.rotation,
	)
	tile_pos.go_relative(dx, dy, 0)
	return tile_pos

"""The distance the player can see."""
_vision_radius = 32
"""The size of a single tile on the screen."""
_tile_size = 16

def draw_tile(world, queue, pos: ChunkPosition, dx, dy):
	"""Add the tile at the given position to the draw queue.
	
	dx and dy are the offset from the player's position in tiles.
	"""
	block = world.get_block(pos._coordinate(),
			pos._local_x(), pos._local_y(), pos._local_z()
	)
	if block is None:
		sprite = "tile"
	else:
		block = Blocky(block)
		sprite = block.appearance
	queue.sprite(
			sprite,
			white,
			_tile_size * (dy + _vision_radius), _tile_size * (_vision_radius - dx))

def draw(world, width, height):
	"""Draw the whole game, with given width and height of the rendered area."""
	queue = DrawQueue()
	
	for dx in range(-_vision_radius, _vision_radius + 1):
		for dy in range(-_vision_radius, _vision_radius + 1):
			if dx**2 + dy**2 > _vision_radius**2:
				continue

			pos = get_pos(world.player, dx, dy)
			pos.go_relative(0, 0, -1)
			draw_tile(world, queue, pos, dx, dy)
			pos.go_relative(0, 0, 1)
			draw_tile(world, queue, pos, dx, dy)

	queue.ellipse(rgb_color(1, 0, 0, 1), position(
			_vision_radius * _tile_size, _vision_radius * _tile_size,
			_tile_size, _tile_size
	))
	
	return world, queue

def input(world, pressed, key):
	"""Called when a key is pressed or released.
	
	pressed is True when the key is pressed, False when released.
	key is an int representing the SDL key code, as in input.keyboard_table.
	"""

	if pressed:
		player_pos = Positioned(world.player)
		if key == keyboard_table.up:
			player_pos.go_forward()
			get_pos.cache_clear()
		if key == keyboard_table.left:
			player_pos.turn_left()
			get_pos.cache_clear()
		if key == keyboard_table.right:
			player_pos.turn_right()
			get_pos.cache_clear()
		if key == keyboard_table.down:
			player_pos.move(-1, 0, 0)
			get_pos.cache_clear()

	return world
