import random

from dbecs.entity import Entity
from hyperbolic.chunk import Chunk
from hyperbolic.coordinate import Coordinate, origin, neighbors
from hyperbolic.position import Positioned, north
from hyperbolic.size import Sized
from hyperbolic.random import seed
from hyperbolic.region import region_options

class World:
	"""Represents the current status of the simulation."""
	def __init__(self):
		self.player = Entity()
		self.player.add_component(Positioned, origin, 0, 0, 1, north)

		"""All chunks are currently stored in this dictionary.

		To get a chunk, you probably want to use the method get_chunk.
		"""
		self._chunk_map = {}

	def place_block(self, block):
		"""Add a sized and positioned object to the world."""
		block_pos = Positioned(block)
		block_size = Sized(block)

		chunk = self.get_chunk(block_pos.coordinate)
		for dx in range(0, block_size.size_x):
			for dy in range(0, block_size.size_y):
				for dz in range(0, block_size.size_z):
					x = block_pos.local_x + dx
					y = block_pos.local_y + dy
					z = block_pos.local_z + dz
					chunk[x, y, z] = block

	def get_block(self, coordinate, x, y, z):
		"""Get the block with given coordinates."""
		return self.get_chunk(coordinate)[x, y, z]

	def _get_range_x(self, coordinate, x, y, z, sx):
		"""Yield sx blocks from the given x upwards."""
		for dx in range(0, sx):
			yield self.get_block(coordinate, x + dx, y, z)

	def _get_range_xy(self, coordinate, x, y, z, sx, sy):
		"""Yield sy rows of blocks from the given y upwards."""
		for dy in range(0, sy):
			yield self._get_range_x(coordinate, x, y + dy, z, sx)

	def get_range(self, coordinate, x, y, z, sx, sy, sz):
		"""Yield sz layers of blocks from the given z upwards."""
		for dz in range(0, sz):
			yield self._get_range_xy(coordinate, x, y, z + dz, sx, sy)

	def _generate_chunk(self, coordinate: Coordinate):
		"""Create a new chunk at the given coordinate.
		
		The contents of the chunk will be determined by the Region this chunk is in.
		"""
		regions = []
		for neighbor in neighbors(coordinate):
			chunk = self.get_maybe_chunk(neighbor)
			if chunk:
				regions.append(chunk.region)
			else:
				regions.append(None)

		# FIXME: hacky workaround for generating random numbers, why?
		random.randint(0, 4)
		with seed(coordinate):
			choice = random.choice(regions)
			if not choice:
				choice = random.choice(region_options)()

			blocks = choice.generate_chunk(coordinate)

		return Chunk(coordinate, blocks, choice)

	def get_chunk(self, coordinate: Coordinate):
		"""Get the chunk at the given coordinate.
		
		If the chunk didn't exist yet, it will be generated.
		"""
		try:
			chunk = self._chunk_map[coordinate]
		except KeyError:
			chunk = self._generate_chunk(coordinate)
			self._chunk_map[coordinate] = chunk
		return chunk

	def get_maybe_chunk(self, coordinate: Coordinate):
		"""Get the chunk at the given coordinate.
		
		If the chunk didn't exist yet, returns None.
		"""
		try:
			return self._chunk_map[coordinate]
		except KeyError:
			return None
